"""Configuration module for drift_monitor detection client."""

import datetime as dt

import jwt
from libmytoken import MytokenServer
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    """Settings class for drift_monitor detection client."""

    model_config = SettingsConfigDict(
        case_sensitive=False,
    )

    DRIFT_MONITOR_URL: str
    DRIFT_MONITOR_API_VERSION: str = "v1"
    DRIFT_MONITOR_TIMEOUT: int = 10

    DRIFT_MONITOR_MYTOKEN: str
    MYTOKEN_SERVER: str = "https://mytok.eu"

    TESTING: bool = False

    @property
    def monitor_url(self) -> str:
        """Get the monitor URL."""
        return f"{self.DRIFT_MONITOR_URL}/api/{self.DRIFT_MONITOR_API_VERSION}"


class AccessToken:
    """Access token class for mytoken server."""

    def __init__(self):
        self._at = MytokenServer(settings.MYTOKEN_SERVER).AccessToken
        self.value = None
        self.decode_options = {
            "algorithms": ["HS256"],
            "options": {"verify_signature": False},
        }
        self.info = {"exp": 0}

    def __call__(self):
        """Get the access token from mytoken server."""
        if self.info["exp"] < dt.datetime.now(dt.timezone.utc).timestamp():
            self.value = self._at.get(settings.DRIFT_MONITOR_MYTOKEN)
            self.info = jwt.decode(self.value, **self.decode_options)
        return self.value


# Initialize the settings object
settings = Settings()
access_token = AccessToken()



"""Test example module."""

import pytest

from drift_monitor import DriftMonitor


@pytest.fixture(scope="function")
def monitor(request_mock, experiment_name, model_id):
    """Create a drift run on the drift monitor server."""
    return DriftMonitor(experiment_name, model_id)


@pytest.mark.parametrize("experiment_name", ["experiment_1"])
@pytest.mark.parametrize("model_id", ["some_model"])
def test_context(monitor):
    """Test the method concept raises out of context error."""
    with pytest.raises(RuntimeError) as excinfo:
        monitor(True, {"threshold": 0.5})
    assert str(excinfo.value) == "Drift monitor context not started."



"""Test module for creation of experiments."""

from unittest import mock

import pytest

from drift_monitor import DriftMonitor, models


@pytest.fixture(scope="function", autouse=True)
def find_mock():
    """Patch requests module with MagicMocks."""
    with mock.patch("drift_monitor.find_experiment") as find:
        yield find


@pytest.fixture(scope="function", autouse=True)
def mocks(request_mock, find_mock, db_experiment, db_drift):
    """Mock the requests module."""
    request_mock.post.return_value = mock.MagicMock(json=db_drift.copy)
    request_mock.put.return_value = mock.MagicMock(json=db_drift.copy)
    find_mock.return_value = models.Experiment(db_experiment)


@pytest.fixture(scope="function")
def monitor(mocks, experiment_name, db_drift):
    """Create a drift run on the drift monitor server."""
    try:
        with DriftMonitor(experiment_name, db_drift["model"]) as _monitor:
            raise ValueError("Failed to create drift run.")
    except ValueError:
        pass
    return _monitor


@pytest.mark.parametrize("experiment_name", ["experiment_1"])
@pytest.mark.parametrize("drift_id", ["00000000-0000-0000-0000-000000000001"])
@pytest.mark.usefixtures("monitor")
def test_request(request_mock):
    """Test the drift run was created on the server."""
    assert request_mock.post.call_count == 1
    assert request_mock.put.call_count == 1


@pytest.mark.parametrize("experiment_name", ["experiment_1"])
@pytest.mark.parametrize("drift_id", ["00000000-0000-0000-0000-000000000001"])
@pytest.mark.usefixtures("monitor")
def test_status(request_mock):
    """Test the drift run was completed on the server."""
    assert request_mock.post.call_args[1]["json"]["job_status"] == "Running"
    assert request_mock.put.call_args[1]["json"]["job_status"] == "Failed"



"""Query functions for drift watch server."""

from typing import List
from uuid import UUID

import marshmallow as ma
import requests

from drift_monitor import models, schemas
from drift_monitor.config import access_token, settings

# Entitlements: API Methods to list, entitlements.


def get_entitlements() -> List[str]:
    """Get the entitlements of the token user."""
    response = requests.get(
        url=f"{settings.monitor_url}/entitlements",
        headers={"Authorization": f"Bearer {access_token()}"},
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()
    return response.json()["items"]


# Experiments: API Methods to list, register, edit and remove experiments.


def search_experiment(query, page=1, page_size=10):
    """Search for experiments on the drift monitor server."""
    response = requests.post(
        url=f"{settings.monitor_url}/experiment/search",
        headers={"Authorization": f"Bearer {access_token()}"},
        params={"page": page, "page_size": page_size},
        json=query,
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()
    return (
        [models.Experiment(data) for data in response.json()],
        response.headers["X-Pagination"],
    )


def post_experiment(attributes) -> models.Experiment:
    """Create a new experiment on the drift monitor server."""
    response = requests.post(
        url=f"{settings.monitor_url}/experiment",
        headers={"Authorization": f"Bearer {access_token()}"},
        json=schemas.CreateExperiment().load(attributes),
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()
    return models.Experiment(response.json())


def get_experiment(experiment_id: UUID) -> models.Experiment:
    """Get an experiment from the drift monitor server."""
    response = requests.get(
        url=f"{settings.monitor_url}/experiment/{experiment_id}",
        headers={"Authorization": f"Bearer {access_token()}"},
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()
    return models.Experiment(response.json())


def put_experiment(experiment: models.Experiment) -> models.Experiment:
    """Update an experiment on the drift monitor server."""
    response = requests.put(
        url=f"{settings.monitor_url}/experiment/{experiment.id}",
        headers={"Authorization": f"Bearer {access_token()}"},
        json=schemas.UpdateExperiment(unknown=ma.EXCLUDE).dump(experiment),
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()
    return models.Experiment(response.json())


def delete_experiment(experiment_id: UUID) -> None:
    """Delete an experiment from the drift monitor server."""
    response = requests.delete(
        url=f"{settings.monitor_url}/experiment/{experiment_id}",
        headers={"Authorization": f"Bearer {access_token()}"},
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()
    return None


# Drifts: API Methods to list, create, complete and fail drift runs.


def search_drift(experiment, query, page=1, page_size=10):
    """Search for drift runs on the drift monitor server."""
    exp_route = f"experiment/{experiment['id']}"
    response = requests.post(
        url=f"{settings.monitor_url}/{exp_route}/drift/search",
        headers={"Authorization": f"Bearer {access_token()}"},
        params={"page": page, "page_size": page_size},
        json=query,
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()
    return (
        [models.Drift(data) for data in response.json()],
        response.headers["X-Pagination"],
    )


def post_drift(experiment, attributes) -> models.Drift:
    """Create a new drift run on the drift monitor server."""
    exp_route = f"experiment/{experiment.id}"
    response = requests.post(
        url=f"{settings.monitor_url}/{exp_route}/drift",
        headers={"Authorization": f"Bearer {access_token()}"},
        json=schemas.CreateDrift().load(attributes),
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()
    return models.Drift(response.json())


def get_drift(experiment, drift_id: UUID) -> models.Drift:
    """Get a drift run from the drift monitor server."""
    exp_route = f"experiment/{experiment.id}"
    response = requests.get(
        url=f"{settings.monitor_url}/{exp_route}/drift/{drift_id}",
        headers={"Authorization": f"Bearer {access_token()}"},
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()
    return models.Drift(response.json())


def put_drift(experiment, drift: models.Drift) -> models.Drift:
    """Update a drift run on the drift monitor server."""
    exp_route = f"experiment/{experiment.id}"
    response = requests.put(
        url=f"{settings.monitor_url}/{exp_route}/drift/{drift.id}",
        headers={"Authorization": f"Bearer {access_token()}"},
        json=schemas.UpdateDrift(unknown=ma.EXCLUDE).dump(drift),
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()
    return models.Drift(response.json())


def delete_drift(experiment, drift_id: UUID) -> None:
    """Delete a drift run from the drift monitor server."""
    exp_route = f"experiment/{experiment['id']}"
    response = requests.delete(
        url=f"{settings.monitor_url}/{exp_route}/drift/{drift_id}",
        headers={"Authorization": f"Bearer {access_token()}"},
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()
    return None


# Users: API Methods to list, register and update users.


def search_user(query, page=1, page_size=10):
    """Search for users on the drift monitor server."""
    response = requests.post(
        url=f"{settings.monitor_url}/user/search",
        headers={"Authorization": f"Bearer {access_token()}"},
        params={"page": page, "page_size": page_size},
        json=query,
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()
    return (
        [models.User(data) for data in response.json()],
        response.headers["X-Pagination"],
    )


def post_user() -> models.User:
    """Create a new user on the drift monitor server."""
    response = requests.post(
        url=f"{settings.monitor_url}/user",
        headers={"Authorization": f"Bearer {access_token()}"},
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()
    return models.User(response.json())


def self_user() -> models.User:
    """Get the token user from the drift monitor server."""
    response = requests.get(
        url=f"{settings.monitor_url}/user/self",
        headers={"Authorization": f"Bearer {access_token()}"},
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()
    return models.User(response.json())


def update_user() -> models.User:
    """Update the token user in the application database."""
    response = requests.put(
        url=f"{settings.monitor_url}/user/self",
        headers={"Authorization": f"Bearer {access_token()}"},
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()
    return models.User(response.json())



"""Dataclasses for drift_monitor module."""

import dataclasses
import uuid
from abc import ABC

import marshmallow as ma

from drift_monitor import schemas


class _BaseModel(ABC):  # pylint: disable=too-few-public-methods
    id: uuid.UUID | None = None
    created_at: str | None = None


@dataclasses.dataclass(init=False)
class Experiment(_BaseModel):
    """Experiment dataclass."""

    name: str
    description: str | None = None
    public: bool = False
    permissions: list = dataclasses.field(default_factory=list)

    def __init__(self, data):
        load = schemas.Experiment(unknown=ma.INCLUDE).load(data)
        self.__dict__.update(load)


@dataclasses.dataclass(init=False)
class Drift(_BaseModel):
    """Drift dataclass."""

    model: str
    schema_version: str | None = None
    job_status: str = "Running"
    tags: list = dataclasses.field(default_factory=list)
    drift_detected: bool = False
    parameters: dict = dataclasses.field(default_factory=dict)

    def __init__(self, data):
        load = schemas.Drift(unknown=ma.INCLUDE).load(data)
        self.__dict__.update(load)


@dataclasses.dataclass(init=False)
class User(_BaseModel):
    """User dataclass."""

    subject: str
    issuer: str
    email: str

    def __init__(self, data):
        load = schemas.User(unknown=ma.INCLUDE).load(data)
        self.__dict__.update(load)



"""Test module for creation of drifts."""

# pylint: disable=redefined-outer-name

from unittest import mock

import pytest

import drift_monitor


@pytest.fixture(scope="function", autouse=True)
def mocks(request_mock, db_experiment):
    """Mock the requests module."""
    request_mock.post.return_value = mock.MagicMock(
        json=lambda: [db_experiment],
        headers={"X-Pagination": "somevalues"},
    )


@pytest.fixture(scope="function")
def experiment(mocks, experiment_name):
    """Create a drift run on the drift monitor server."""
    return drift_monitor.find_experiment(experiment_name)


@pytest.mark.parametrize("experiment_name", ["experiment_1"])
@pytest.mark.usefixtures("experiment")
def test_request(request_mock):
    """Test the drift run was created on the server."""
    assert request_mock.post.call_count == 1


@pytest.mark.parametrize("experiment_name", ["experiment_1"])
def test_returns(experiment, db_experiment):
    """Test correct return values from new experiment"""
    assert experiment.__dict__ == db_experiment



""""This module contains the schema for the API request and response."""

import marshmallow as ma
from marshmallow import validate


class _BaseReqSchema(ma.Schema):
    pass


class _BaseRespSchema(ma.Schema):
    _id = ma.fields.UUID(required=True, data_key="id", dump_only=True)
    created_at = ma.fields.String(required=True, dump_only=True)


class Entitlements(ma.Schema):
    """
    Entitlement is a string of the form "vo#role".
    """

    items = ma.fields.List(ma.fields.String(), required=True)


class Permission(ma.Schema):
    """
    Permissions defines the entity and the access level.
    """

    class Meta:  # pylint: disable=R0903, C0115
        unknown = ma.RAISE

    entity = ma.fields.String(required=True)
    level = ma.fields.String(required=True)


class _BaseExperiment(ma.Schema):
    name = ma.fields.String(required=True)
    description = ma.fields.String()
    public = ma.fields.Boolean(load_default=False)
    permissions = ma.fields.List(
        ma.fields.Nested(Permission),
        load_default=[],
    )


class Experiment(_BaseExperiment, _BaseRespSchema):
    """
    Experiment is a pointer to the collection of drifts.
    Authentication is carried by the groups that point here.
    A name is required for easy identification.
    Includes the list of permissions for the groups.
    """


class CreateExperiment(_BaseExperiment, _BaseReqSchema):
    """Create Experiment Schema."""


class UpdateExperiment(_BaseExperiment, _BaseReqSchema):
    """Update Experiment Schema."""


class User(_BaseRespSchema):
    """
    User represent a person that access and uses the API.
    User's ids can also be used as group_ids.
    """

    subject = ma.fields.String(dump_only=True)
    issuer = ma.fields.String(dump_only=True)
    email = ma.fields.Email(dump_only=True)


class UsersEmails(ma.Schema):
    """Schema for a list of emails."""

    emails = ma.fields.List(
        ma.fields.Email(),
        required=True,
        validate=validate.Length(min=1, max=10),
    )


class UsersIds(ma.Schema):
    """Schema for a list of ids."""

    ids = ma.fields.List(
        ma.fields.UUID(),
        required=True,
        validate=validate.Length(min=1, max=10),
    )


class BaseDrift(ma.Schema):
    """Base drift schema"""

    drift = ma.fields.Bool(required=True)
    parameters = ma.fields.Dict()


tag = ma.fields.String(validate=validate.Length(min=1, max=20))
status_options = validate.OneOf(["Running", "Completed", "Failed"])


class _BaseDriftJob(ma.Schema):
    job_status = ma.fields.String(required=True, validate=status_options)
    tags = ma.fields.List(tag, load_default=[])
    model = ma.fields.String(required=True)
    drift_detected = ma.fields.Bool(load_default=False)
    parameters = ma.fields.Dict(load_default={})


class Drift(_BaseDriftJob, _BaseRespSchema):
    """
    Response Job Schema. A drift is the basic unit of the API.
    It contains the drift information for a specific model and job.
    """

    schema_version = ma.fields.String(required=True, dump_only=True)


class CreateDrift(_BaseDriftJob, _BaseReqSchema):
    """Create Job Schema for job."""


class UpdateDrift(_BaseDriftJob, _BaseReqSchema):
    """Update Job Schema for job."""



"""Drift Monitor Client package.
This package contains the client code for the drift monitor service.
"""

from typing import Any, Dict, List, Optional

import requests

from drift_monitor import models, queries, utils


class DriftMonitor:
    """Drift Monitor context.
    This class is a context manager for the drift monitor service. It is used
    as an abstraction for the user to interact with the drift monitor service.

    When the context is entered, the drift monitor sends a POST request to the
    server to create a drift run. When the context is exited, the drift monitor
    sends a PUT request to the server to complete the drift run.

    Args:
        experiment (str): The name of the experiment.
        model_id (str): The model ID to monitor.
        tags (list, optional): The tags for the drift. Defaults to None.

    Example:
        >>> with DriftMonitor("experiment_1", "model_1") as monitor:
        ...    hypothesis_result, parameters = concept_detector()
        ...    monitor(detected, detection_parameters)
    """

    def __init__(
        self,
        experiment_name: str,
        model_id: str,
        tags: Optional[List[str]] = None,
    ):
        self._experiment_name: str = experiment_name
        self._experiment: models.Experiment | None = None
        self._attributes = {"model": model_id, "tags": tags or []}
        self._drift: models.Drift | None = None

    def __enter__(self) -> "DriftMonitor":
        self._experiment = find_experiment(self._experiment_name)
        attributes = {"job_status": "Running", **self._attributes}
        self._drift = queries.post_drift(self._experiment, attributes)
        return self

    def __call__(
        self,
        detected: bool,
        parameters: Dict[str, Any],
    ) -> None:
        """Prepare drift detection results for transmission to server.

        Args:
            detected (bool): Whether concept drift was detected.
            parameters (dict): The parameter values from detection.

        Raises:
            RuntimeError: If the drift monitor context is not started.
        """
        if self._experiment is None:
            raise RuntimeError("Drift monitor context not started.")
        if self._drift is None:
            raise RuntimeError("Drift removed while context active.")
        parameters = utils.convert_to_serializable(parameters)
        self._drift.drift_detected = bool(detected)  # Ensure serialization
        self._drift.parameters = parameters

    def __exit__(
        self,
        exc_type: Optional[type],
        _exc_value: Optional[BaseException],
        _traceback: Optional[Any],
    ) -> None:
        if self._experiment is None:
            raise RuntimeError("Drift monitor context not started.")
        if self._drift is None:
            raise RuntimeError("Drift removed while context active.")
        if exc_type:
            self._drift.job_status = "Failed"  # New status
            self._drift = queries.put_drift(self._experiment, self._drift)
        else:
            self._drift.job_status = "Completed"
            self._drift = queries.put_drift(self._experiment, self._drift)
        self._experiment = None  # Reset drift object


def register(accept_terms: bool = False) -> None:
    """Registers the token user in the application database.
    By using this function, you accept that the user derived from the token
    will be registered in the application database and agree to the terms of
    service.

    Args:
        accept_terms (bool, optional): Whether to accept the terms of service.
            Defaults to False.

    Raises:
        ValueError: If the user is already registered or terms are not accepted.
    """
    if not accept_terms:
        raise ValueError("You must accept the terms of service.")
    try:
        queries.post_user()
    except requests.HTTPError as error:
        if error.response.status_code == 409:
            queries.update_user()
            return  # User already registered
        raise error


def find_experiment(experiment_name: str) -> models.Experiment:
    """Get an experiment from the drift monitor server.

    Args:
        experiment_name (str): The name of the experiment.

    Returns:
        dict: The experiment object or None if not found.

    Raises:
        ValueError: If the experiment is not found.
    """
    search_query = {"name": experiment_name}
    experiments, _ = queries.search_experiment(search_query)
    if experiments is []:
        raise ValueError("Experiment not found.")
    return experiments[0]  # Return first result


def new_experiment(
    name: str,
    description: str,
    public: bool = False,
    permissions: Optional[Dict[str, Any]] = None,
) -> models.Experiment:
    """Create a new experiment in the drift monitor service.

    Args:
        name (str): The name of the experiment.
        description (str): The description of the experiment.
        public (bool, optional): Whether the experiment is public.
            Defaults to False.
        permissions (dict, optional): The permissions for the experiment.
            Defaults to None.

    Returns:
        dict: The experiment object.
    """
    try:
        return queries.post_experiment(
            {
                "name": name,
                "description": description,
                "public": public,
                "permissions": permissions if permissions else [],
            }
        )
    except requests.HTTPError as error:
        if error.response.status_code == 409:
            raise ValueError("Experiment already exists.") from error
        raise error
